"""

    File:    DbLoadingApi.py

    Providing an API for WFM to parse the cif file and run MYSQL commands to load data to the database 'da_internal'.
 
   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "Oct 4, 2013"

 Updates:

   07-Feb-2014  jdw   Add socket support
   21-May-2015  LC Add function doDataLoadingBcp()
"""

import os,sys,traceback
from subprocess import Popen, PIPE
#
from wwpdb.utils.config.ConfigInfo import ConfigInfo

import signal

class DbLoadingApi(object):
    """
    """
    def __init__(self, log=sys.stderr, verbose=False):
        """
           
        """
        self.__lfh = log
        self.__verbose = verbose
        self.__debug=True
        cI=ConfigInfo()
        self.__pyPath      = cI.get("SITE_PYTHON_SOURCE_PATH")
        self.__pkgPath     = cI.get("SITE_PACKAGES_PATH")
        self.__resrcPath   = cI.get("SITE_RESOURCE_DIRECTORY_PATH")
        self.__dbServer    = cI.get("SITE_DB_SERVER")
        #self.__dbHost      = "pdb-b-linux-6.rutgers.edu"
        self.__dbHost      = cI.get("SITE_DB_HOST_NAME")
        self.__dbUser      = cI.get("SITE_DB_USER_NAME")
        self.__dbPw        = cI.get("SITE_DB_PASSWORD")
        self.__dbPort      = str(cI.get("SITE_DB_PORT_NUMBER"))
        self.__dbSocket    = cI.get("SITE_DB_SOCKET")
        self.__archivePath  = cI.get("SITE_ARCHIVE_STORAGE_PATH")
        
        self.__dbName    = "da_internal"
        self.__workPath  = os.path.join(self.__archivePath,"archive")
        self.__mysql  = "/usr/bin/mysql "
        self.__dbLoader  = os.path.join(self.__pkgPath,"dbloader","bin","db-loader") 
        
        self.__mapping   = os.path.join(self.__resrcPath,"status_rcsb_schema_da.cif")
        
        self.__workDir   = "dbdata"                                 
    
    def doDataLoading(self, depId, sessionDir):
        """
           Take deposition id and session directory as input

        """
        """
          If a sequence of commands appears in a pipeline, and one of the
          reading commands finishes before the writer has finished, the
          writer receives a SIGPIPE signal.
          Set signal for running os.system() with broken PIPE problem
        """
        #signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        
        depId = depId.upper()
        cifPath = self.__workPath+"/"+depId+"/"
        ## test if there is any cif file if yes do ...
        
        if(os.path.exists(cifPath + depId+'_model_P1.cif.V1')):
            dataDir = sessionDir+"/"+self.__workDir
            log1 = "db-loader.log"
            if not os.path.exists(dataDir):
                print("Creating "+dataDir)
                os.makedirs(dataDir)
           
            cmd = "cd "+dataDir
            cmd += "; rm -f *"
            cmd +="; ls -tl "+cifPath+"D_*model_P1.cif.V[1-9]* | head -n1 | awk '{print $9}' >FILELIST"
            cmd +="; "+self.__dbLoader+" -server mysql -list FILELIST -map "+self.__mapping+" -db "+self.__dbName +" >& "+log1
            os.system(cmd)
            #print cmd
            #for file in os.listdir(dataDir):
            #    print file
            
            if(os.path.exists(dataDir+"/FILELIST")):
                # checking generated files
                file1 = os.path.join(dataDir,"DB_LOADER.sql")
                log2  = os.path.join(dataDir,"data_loading.log")
                if(os.path.exists(file1)):
                    cmd = "cd "+dataDir
                    if self.__dbSocket is None:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file1+" >&"+log2
                    else:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file1+" >&"+log2
                    
                    os.system(cmd)
                    if(os.path.exists(log2)):
                        print("Finished the database commands")
                        file = open(log2, 'r')
                        for line in file:
                            for word in line.split():
                                if(word.upper()=="ERROR"):
                                    print("DbLoadingApi::doDataLoading(): ERROR found during the database loading. Please check the log file "+log2+" for details.")
                                    
                else:
                    print("DbLoadingApi::doDataLoading(): db-loader didn't generate the data file \"DB_LOADER.sql\". Please check the log file "+os.path.join(dataDir,log1)+" for details.")
            else:
                print("DbLoadingApi::doDataLoading(): No cif file found. Please check if the cif file exists.")
                    
            
        else:
            print("DbLoadingApi::doDataLoading(): No any cif file found.")

    def doLoadStatus(self, pdbxFilePath, sessionDir):
        """
           Load the input file into the status database and session directory as input

        """
        #signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        self.__lfh.write("DbLoadingApi::doLoadStatus(): starting with file %s and session %s\n" % (pdbxFilePath,sessionDir))
        try:
            if(os.path.exists(pdbxFilePath)):
                dataDir = sessionDir+"/"+self.__workDir
                log1 = "db-loader.log"
                if not os.path.exists(dataDir):
                    print("Creating "+dataDir)
                    os.makedirs(dataDir)

                cmd = "cd " + dataDir
                cmd +="; "+self.__dbLoader+" -server mysql -f " + pdbxFilePath + " -map " + self.__mapping + " -db " + self.__dbName +" >& "+log1
                os.system(cmd)
                if self.__debug:
                    self.__lfh.write("DbLoadingApi::doLoadStatus(): db-loader command %s\n" % cmd)

                file1 = os.path.join(dataDir,"DB_LOADER.sql")
                log2  = os.path.join(dataDir,"status_load.log")            
                if(os.path.exists(file1)):
                    cmd = "cd "+dataDir
                    if self.__dbSocket is None:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file1+" >& "+log2
                    else:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file1+" >& "+log2
                    if self.__debug:
                        self.__lfh.write("DbLoadingApi::doLoadStatus(): database server command %s\n" % cmd)
                    os.system(cmd)
                    if(os.path.exists(log2)):
                        file = open(log2, 'r')
                        for line in file:
                            for word in line.split():
                                if(word.upper()=="ERROR"):
                                    self.__lfh.write("DbLoadingApi::doLoadStatus(): ERROR found during the database loading. Please check the log file "+log2+" for details.\n")
                                    return False
                    else:
                        self.__lfh.write("DbLoadingApi::doLoadStatus(): db-loader didn't generate the data file \"DB_LOADER.sql\". Please check the log file "+os.path.join(dataDir,log1)+" for details.\n")

                    #
                    self.__lfh.write("DbLoadingApi::doLoadStatus(): completed\n")
                    return True
                else:
                    self.__lfh.write("DbLoadingApi::doLoadStatus(): failing, no load file created.\n")
                    return False
            else:
                self.__lfh.write("DbLoadingApi::doLoadStatus(): failing, no input cif file found.\n")
                return False
        except:
            self.__lfh.write("DbLoadingApi::doLoadStatus(): failing, with exception.\n")            
            traceback.print_exc(file=self.__lfh)                                        

    def doDataLoadingBcp(self, depId, sessionDir):
        """
           Similar as doDataLoading(), Run db-loader with the option to
           get bcp data files, not sql commands.

        """
        """
          If a sequence of commands appears in a pipeline, and one of the
          reading commands finishes before the writer has finished, the
          writer receives a SIGPIPE signal.
          Set signal for running os.system() with broken PIPE problem
        """
        #signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        #print self.__dbHost
        
        depId = depId.upper()
        cifPath = self.__workPath+"/"+depId+"/"
        ## test if there is any cif file if yes do ...
        
        if(os.path.exists(cifPath + depId+'_model_P1.cif.V1')):
            dataDir = sessionDir+"/"+self.__workDir
            log1 = "db-loader.log"
            if not os.path.exists(dataDir):
                print("Creating "+dataDir)
                os.makedirs(dataDir)
            
            cmd = "cd "+dataDir
            cmd += "; rm -f *"
            cmd +="; ls -tl "+cifPath+"D_*model_P1.cif.V[1-9]* | head -n1 | awk '{print $9}' >FILELIST"
            cmd +="; "+self.__dbLoader+" -server mysql -list FILELIST -map "+self.__mapping+" -db "+self.__dbName +" -bcp >& "+log1
            #print cmd 
            os.system(cmd)
            
            #for file in os.listdir(dataDir):
            #    print file
            
            if(os.path.exists(dataDir+"/FILELIST")):
                # checking generated files
                file1 = os.path.join(dataDir,"DB_LOADER_DELETE.sql")
                file2 = os.path.join(dataDir,"DB_LOADER_LOAD.sql")
                log2  = os.path.join(dataDir,"data_loading.log")
                log3  = os.path.join(dataDir,"data_delete.log")
                if(os.path.exists(file2)):
                    cmd = "cd "+dataDir
                    if self.__dbSocket is None:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file1+">& "+log3
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file2+">& "+log2
                        
                    else:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file1+">& "+log3
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file2+">& "+log2
                    #print cmd
                    os.system(cmd)
                    if(os.path.exists(log2)):
                        print("Finished the database commands")
                        file = open(log2, 'r')
                        for line in file:
                            for word in line.split():
                                if(word.upper()=="ERROR"):
                                    print("DbLoadingApi::doDataLoading(): ERROR found during the database loading. Please check the log file "+log2+" for details.")
                                    
                else:
                    print("DbLoadingApi::doDataLoading(): db-loader didn't generate the data file \"DB_LOADER_LOAD.sql\". Please check the log file "+os.path.join(dataDir,log1)+" for details.")
            else:
                print("DbLoadingApi::doDataLoading(): No cif file found. Please check if the cif file exists.")
                    
            
        else:
            print("DbLoadingApi::doDataLoading(): No any cif file found.")

    def doDataLoadingByMapping(self, depId, sessionDir, mappingFile, dbName):
        """
           Similar as doDataLoading(), Run db-loader with the option to
           get bcp data files using the giving mapping file and dbname.

        """
        """
          If a sequence of commands appears in a pipeline, and one of the
          reading commands finishes before the writer has finished, the
          writer receives a SIGPIPE signal.
          Set signal for running os.system() with broken PIPE problem
        """
        #signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        #print self.__dbHost
        
        depId = depId.upper()
        myMappingFile = os.path.join(self.__resrcPath, mappingFile)
        cifPath = self.__workPath+"/"+depId+"/"
        ## test if there is any cif file if yes do ...
        
        if(os.path.exists(cifPath + depId+'_model_P1.cif.V1')):
            dataDir = sessionDir+"/"+self.__workDir
            log1 = "db-loader.log"
            if not os.path.exists(dataDir):
                print("Creating "+dataDir)
                os.makedirs(dataDir)
            
            cmd = "cd "+dataDir
            cmd += "; rm -f *"
            cmd +="; ls -tl "+cifPath+"D_*model_P1.cif.V[1-9]* | head -n1 | awk '{print $9}' >FILELIST"
            cmd +="; "+self.__dbLoader+" -server mysql -list FILELIST -map "+myMappingFile+" -db "+dbName +" -bcp >& "+log1
            #print cmd 
            os.system(cmd)
            
            #for file in os.listdir(dataDir):
            #    print file
            
            if(os.path.exists(dataDir+"/FILELIST")):
                # checking generated files
                file1 = os.path.join(dataDir,"DB_LOADER_DELETE.sql")
                file2 = os.path.join(dataDir,"DB_LOADER_LOAD.sql")
                log2  = os.path.join(dataDir,"data_loading.log")
                log3  = os.path.join(dataDir,"data_delete.log")
                if(os.path.exists(file2)):
                    cmd = "cd "+dataDir
                    if self.__dbSocket is None:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file1+">& "+log3
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" <"+file2+">& "+log2
                        
                    else:
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file1+">& "+log3
                        cmd += "; "+self.__mysql+"-u "+self.__dbUser+" -p"+self.__dbPw+" -h "+self.__dbHost+" -P "+self.__dbPort+" -S "+self.__dbSocket+ " <"+file2+">& "+log2
                    #print cmd
                    os.system(cmd)
                    if(os.path.exists(log2)):
                        print("Finished the database commands")
                        file = open(log2, 'r')
                        for line in file:
                            for word in line.split():
                                if(word.upper()=="ERROR"):
                                    print("DbLoadingApi::doDataLoading(): ERROR found during the database loading. Please check the log file "+log2+" for details.")
                                    
                else:
                    print("DbLoadingApi::doDataLoading(): db-loader didn't generate the data file \"DB_LOADER_LOAD.sql\". Please check the log file "+os.path.join(dataDir,log1)+" for details.")
            else:
                print("DbLoadingApi::doDataLoading(): No cif file found. Please check if the cif file exists.")
                    
            
        else:
            print("DbLoadingApi::doDataLoading(): No any cif file found.")

                
    if __name__ == '__main__':
        pass
    



##
# File:  StatusLoadWrapper.py
# Date:  18-May-2015
#
# Updates:
##
"""
Wrapper for simple database loader for da_internal database --

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
import os
import traceback

from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.io.locator.PathInfo import PathInfo
from wwpdb.utils.wf.dbapi.DbLoadingApi import DbLoadingApi


class StatusLoadWrapper(object):

    """ Update release status items.

    """

    def __init__(self, siteId, verbose=False, log=sys.stderr):
        """
         :param `verbose`:  boolean flag to activate verbose logging.
         :param `log`:      stream for logging.

        """
        self.__verbose = verbose
        self.__lfh = log
        self.__debug = True
        self.__siteId = siteId
        #
        self.__cI = ConfigInfo(self.__siteId)
        self.__pI = PathInfo(siteId=self.__siteId, sessionPath='.', verbose=self.__verbose, log=self.__lfh)

    def dbLoad(self, depSetId, fileSource='deposit', versionId='latest', mileStone='deposit'):
        try:
            self.__lfh.write("+StatusLoadWrapper.dbload() site %s loading data set %s %s %s %s\n" % (self.__siteId, depSetId, fileSource, mileStone, versionId))
            pdbxFilePath = self.__pI.getModelPdbxFilePath(dataSetId=depSetId, fileSource=fileSource, versionId=versionId, mileStone=mileStone)
            fD, fN = os.path.split(pdbxFilePath)
            dbLd = DbLoadingApi(log=self.__lfh, verbose=self.__verbose)
            return dbLd.doLoadStatus(pdbxFilePath, fD)
        except:
            if (self.__verbose):
                self.__lfh.write("+StatusLoadWrapper.dbload() dbload failed for %s\n" % depSetId)
                traceback.print_exc(file=self.__lfh)
            return False



import time
import datetime
from decimal import Decimal

# zeroDate is arbitarily 1st Jan 2000 : gives 60 year
zeroDate = datetime.datetime(2000, 1, 1, 0, 0, 0)

def getTimeZero():

   return zeroDate
    
def getTimeNow():

   tNow = datetime.datetime.utcnow()

   c = tNow - zeroDate
   microSec =  ((c.days * 24 * 60 * 60 + c.seconds) * 1000000 +  c.microseconds) / 1000000.0
  

# must use Decimal(repr(time)) as python in its wisdom will only return 3dp if you are lucky
   return Decimal(repr(microSec))

def getTimeSeconds(timestamp):

   zero = time.mktime(zeroDate.timetuple())

   d = float(zero) + float(timestamp)
   return d

def getTimeString(timestamp):

   zero = time.mktime(zeroDate.timetuple())

   fmt = '%Y-%m-%d %H:%M:%S %Z'

   d = float(zero) + float(timestamp)
   return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)

def getTimeFromEpoc(timestamp):

   fmt = '%Y-%m-%d %H:%M:%S %Z'

   d =  float(timestamp)
   return datetime.datetime.fromtimestamp(int(d)).strftime(fmt)

def getTimeReadable(timestamp):

   one  = ((zeroDate.days * 24 * 60 * 60 + zeroDate.seconds) * 1000000 +  zeroDate.microseconds) / 1000000.0
   two = float(str(timestamp))
   return datetime.datetime.utcfromtimestamp(one + two)
   





"""

    File:    LocalDbApi.py

    Providing addintaional APIs for WFE to get info from local database
 
   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "Jun 4, 2010"

     
"""

import os,sys
import MySQLdb
#
from wwpdb.utils.wf.dbapi.DbConnection import DbConnection    
from wwpdb.utils.wf.schema.WfSchemaMap  import WfSchemaMap
from wwpdb.utils.config.ConfigInfo import ConfigInfo

class LocalDbApi(object):
    """
    """
    def __init__(self, log=sys.stderr, verbose=False):
        """
           connect to local database
        """
        self.__lfh = log
        self.__verbose = verbose
        self.__dbServer  = "mysql"
        self.__dbHost    = "pdb-a-linux-5.rutgers.edu"
        self.__dbName    = "msd1"
        self.__dbUser    = "rcsbuser"
        self.__dbPw      = "rcsb0000"
        self.__dbPort    = 3306
        
        self.__myDb  = DbConnection(dbServer=self.__dbServer,dbHost=\
                                    self.__dbHost,dbName=self.__dbName,\
                                    dbUser=self.__dbUser,dbPw=self.__dbPw,\
                                    dbPort=self.__dbPort)
        
        self.__dbcon = self.__myDb.connect()          
        
    
    def getNewDepositedIds(self, interval):
        """
           Get newly deposited ids for RCSB site

           "interval" is a number of days from current time when the python
           code is run.
           
        """
        query = "select lcase(structure_id) from rcsb_status where initial_deposition_date>=DATE_SUB(curdate(), interval %s" % interval +" day) order by structure_id"
        returnList = []
        try:
            self.__dbcon.commit()
            curs = self.__dbcon.cursor()        
            curs.execute(query)
            while True:
                result = curs.fetchone()
                if (result is not None):
                    returnList.append(result[0])
                else:
                    break
        except MySQLdb.Error as e:
            self.__lfh.write("Database error %d: %s\n" % (e.args[0], e.args[1]))
            curs.close()
            self.__dbcon.close()
            sys.exit (1)

        return returnList



"""
   File:    WfSchemaMap.py
   A data class containing schema definitions for WF Database.

   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"
   
"""

class WfSchemaMap:
    _schemaMap = { 
        "DEPOSITION" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"              : "dep_set_id",
                        "PDB_ID"                  : "pdb_id",
                        "INITIAL_DEPOSITION_DATE" : "initial_deposition_date",
                        "ANNOTATOR_INITIALS"      : "annotator_initials",
                        "DEPOSIT_SITE"            : "deposit_site",
                        "PROCESS_SITE"            : "process_site",
                        "STATUS_CODE"             : "status_code",
                        "AUTHOR_RELEASE_STATUS_CODE" : "author_release_status_code",
                        "TITLE"                   : "title",
                        "AUTHOR_LIST"             : "author_list",
                        "EXP_METHOD"              : "exp_method",
                        "STATUS_CODE_EXP"         : "status_code_exp",
                        "SG_CENTER"               : "sg_center",
                        "TITLE_EMDB"              : "title_emdb",
                        "AUTHOR_LIST_EMDB"        : "author_list_emdb",
                        "DEP_AUTHOR_RELEASE_STATUS_CODE_EMDB" : "dep_author_release_status_code_emdb",
                        "STATUS_CODE_EMDB"        : "status_code_emdb",
                        "DATE_BEGIN_PROCESSING"   : "date_begin_processing",
                        "DATE_END_PROCESSING"     : "date_end_processing",
                        "DEPPW"                   : "depPW",
                        "NOTIFY"                  : "notify",
                        "EMAIL"                   : "email",
                        "LOCKING"                 : "locking",
                        "COUNTRY"                 : "country",
                        "NMOLECULE"               : "nmolecule",
                        "EMDB_ID"                 : "emdb_id",
                        "BMRB_ID"                 : "bmrb_id",
                        "STATUS_CODE_BMRB"        : "status_code_bmrb",
                        "STATUS_CODE_OTHER"       : "status_code_other"},
                "TABLE_NAME"  : "deposition"},
        "DATABASE_REF" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"              : "dep_set_id",
                        "DATABASE_NAME"           : "database_name",
                        "DATABASE_CODE"           : "database_code"},
                "TABLE_NAME"  : "database_ref"},
        "WF_TASK" : {
                "ATTRIBUTES" :  {
                        "ORDINAL"             : "ordinal",
                        "WF_TASK_ID"          : "wf_task_id",
                        "WF_INST_ID"          : "wf_inst_id",
                        "WF_CLASS_ID"         : "wf_class_id",
                        "DEP_SET_ID"          : "dep_set_id",
                        "TASK_NAME"           : "task_name",
                        "TASK_STATUS"         : "task_status",
                        "STATUS_TIMESTAMP"    : "status_timestamp",
                        "TASK_TYPE"           : "task_type"},
                "TABLE_NAME"  : "wf_task"},
        "WF_INSTANCE" : {
                "ATTRIBUTES" :  {
                        "ORDINAL"             : "ordinal",
                        "WF_INST_ID"          : "wf_inst_id",
                        "WF_CLASS_ID"         : "wf_class_id",
                        "DEP_SET_ID"          : "dep_set_id",
                        "OWNER"               : "owner",
                        "INST_STATUS"         : "inst_status",
                        "STATUS_TIMESTAMP"    : "status_timestamp"},
                "TABLE_NAME"  : "wf_instance"},
        "WF_REFERENCE" : {
                "ATTRIBUTES" :  {
                        "DEP_SET_ID"          : "dep_set_id",
                        "WF_INST_ID"          : "wf_inst_id",
                        "WF_TASK_ID"          : "wf_task_id",
                        "WF_CLASS_ID"         : "wf_class_id",
                        "HASH_ID"             : "hash_id",
                        "VALUE"               : "value"},
                "TABLE_NAME"  : "wf_reference"},
        "WF_CLASS_DICT" : {
                "ATTRIBUTES" :  {
                        "WF_CLASS_ID"         : "wf_class_id",
                        "WF_CLASS_NAME"       : "wf_class_name",
                        "WF_CLASS_FILE"       : "class_file",
                        "TITLE"               : "title",
                        "AUTHOR"              : "author",
                        "VERSION"             : "version"},
                "TABLE_NAME"  : "wf_class_dict"},
        "PROCESS_INFORMATION" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"        : "dep_set_id",
                        "SERIAL_NUMBER"     : "serial_number",
                        "PROCESS_BEGIN"     : "process_begin",
                        "PROCESS_END"       : "process_end",
                        "REMARK"            : "remark"},
                "TABLE_NAME"  : "process_information"},
        "DA_USERS" : {
                "ATTRIBUTES"  : {
                        "USER_NAME"        : "user_name",
                        "PASSWORD"         : "password",
                        "GROUPNAME"        : "groupname",
                        "EMAIL"            : "email",
                        "INITIALS"         : "initials"},
                "TABLE_NAME"  : "da_users"},
        "DATABASE_RELATED" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"        : "dep_set_id",
                        "DB_NAME"           : "db_name",
                        "DETAILS"           : "details",
                        "CONTENT_TYPE"      : "content_type",
                        "DB_ID"             : "db_id"},
                "TABLE_NAME"  : "database_related"},
        "DATABASE_PDB_OBS_SPR" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"       : "dep_set_id",
                        "ID"               : "id",
                        "DATE"             : "date",
                        "PDB_ID"           : "pdb_id",
                        "REPLACE_PDB_ID"   : "replace_pdb_id"},
                "TABLE_NAME"  : "database_PDB_obs_spr"},
        "AUTHOR_CORRECTIONS" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"        : "dep_set_id",
                        "CORRECTIONS"       : "corrections",
                        "SENDING_DATE"      : "sending_date",
                        "REMARK"            : "content_type"},
                "TABLE_NAME"  : "author_corrections"},
        "DEP_WITH_PROBLEMS" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"        : "dep_set_id",
                        "PROBLEM_TYPE"      : "problem_type",
                        "PROBLEM_DETAILS"   : "problem_details"},
                "TABLE_NAME"  : "dep_with_problems"},
        "RELEASE_REQUEST" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"      : "dep_set_id",
                        "REQ_CITATION"        : "req_citation",
                        "RELEASE_DATE"    : "release_date",
                        "PUBMED_ID"       : "PubMed_id"},
                "TABLE_NAME"  : "release_request"},
        "CONTACT_AUTHOR" : {
                "ATTRIBUTES"  : {
                        "DEP_SET_ID"        : "dep_set_id",
                        "CITATION"          : "name_salutation",
                        "NAME_FIRST"        : "name_first",
                        "NAME_LAST"         : "name_last",
                        "NAME_MI"           : "name_mi",
                        "ROLE"              : "role",
                        "EMAIL"             : "email",
                        "ADDRESS_1"         : "address_1",
                        "ADDRESS_2"         : "address_2",
                        "ADDRESS_3"         : "address_3",
                        "CITY"              : "city",
                        "STATE_PROVINCE"    : "state_province",
                        "POSTAL_CODE"       : "postal_code",
                        "COUNTRY"           : "country",
                        "PHONE"             : "phone",
                        "FAX"               : "fax",
                        "ORGANIZATION_TYPE" : "organization_type"},
                "TABLE_NAME"  : "release_request"},
        "SITE" : {
                "ATTRIBUTES"  : {
                        "CODE"           : "code",
                        "VERBOSE_NAME"   : "verbose_name"},
                "TABLE_NAME"  : "site"},
        "DA_GROUP" : {
                "ATTRIBUTES"  : {
                        "CODE"           : "code",
                        "GROUPNAME"      : "groupname",
                        "SITE"           : "site",
                        "MAIN_PAGE"      : "main_page"},
                "TABLE_NAME"  : "da_group"},
        "SGCENTERS" : {
                "ATTRIBUTES"  : {
                        "CODE"           : "code",
                        "VERBOSE_NAME"   : "verbose_name"},
                "TABLE_NAME"  : "sgcenters"},
                
        "COMMUNICATION" : {
                "ATTRIBUTES"  : {
                        "ORDINAL"           : "ordinal",
                        "SENDER"            : "sender",
                        "RECEIVER"          : "receiver",
                        "DEP_SET_ID"        : "dep_set_id", 
                        "WF_CLASS_ID"       : "wf_class_id",
                        "WF_INST_ID"        : "wf_inst_id",
                        "WF_CLASS_FILE"     : "wf_class_file",
                        "COMMAND"           : "command",
                        "STATUS"            : "status",
                        "STATUS_TIMESTAMP"  : "status_timestamp", 
                        "PARENT_DEP_SET_ID" : "parent_dep_set_id", 
                        "PARENT_WF_CLASS_ID": "parent_wf_class_id",
                        "PARENT_WF_INST_ID" : "parent_wf_inst_id",
                                            
                        },
                "TABLE_NAME"  : "communication"},
        "ENGINE_MONITORING" : {
                "ATTRIBUTES"  : {
                        "CPU_ID"                : "cpu_id",
                        "TOTAL_MEM"             : "total_mem",
                        "TOTAL_PHYSICAL_MEM"    : "total_physical_mem",
                        "TOTAL_VIRTUAL_MEM"     : "total_virtual_mem", 
                        "TOTAL_MEM_USAGE"       : "total_mem_usage",
                        "PHYSICAL_MEM_USAGE"    : "physical_mem_usage",
                        "VIRTUAL_MEM_USAGE"     : "virtual_mem_usage",
                        "CPU_PROCESSES"         : "cpu_processes",
                        "IDS_SET"               : "ids_set",                      
                        },
                "TABLE_NAME"  : "engine_monitoring"},                
       
        }

    # following lists are used in the WfDbApi
    
    _columnForStatus = ['STATUS_CODE','INST_STATUS','TASK_STATUS']
    _objectTables = ['DEPOSITION','WF_CLASS_DICT','WF_INSTANCE','WF_TASK','WF_REFERENCE','DA_USERS']
    _tables = ['DEPOSITION',
               'WF_CLASS_DICT',
               'WF_INSTANCE',
               'WF_TASK',
               'WF_REFERENCE',
               'DA_USERS',
               'DATABASE_PDB_OBS_SPR',
               'DATABASE_REF',
               'DATABASE_RELATED',
               'AUTHOR_CORRECTIONS',
               'RELEASE_REQUEST',
               'DEP_WITH_PROBLEMS',
               'CONTACT_AUTHOR',
               'PROCESS_INFORMATION',
               'SITE',
               'DA_GROUP',
               'SGCENTERS', 
               'COMMUNICATION', 
               'ENGINE_MONITORING',
               ]
    _usefulItems = ['STATUS_CODE',
                    'INST_STATUS',
                    'TASK_STATUS',
                    'DATABASE_CODE',
                    'REPLACE_PDB_ID',
                    'ID',
                    'CONTENT_TYPE',
                    'DB_ID',
                    'DB_NAME',
                    'RELATIONSHIP',
                    'ASSOCIATED_IDS',
                    'ASSESSION_CODE',
                    'CORRECTIONS',
                    'REQ_CITATION',
                    'PROBLEM_TYPE',
                    'PROBLEM_DETAILS',
                    'PUBMED_ID'
                    ]
    _objIds = ['DEP_SET_ID',
               'WF_CLASS_ID',
               'WF_INST_ID',
               'WF_TASK_ID']
    _referencePairs = ['HASH_ID','VALUE']
    _userInfo = ['USER_NAME',
                 'PASSWORD',
                 'GROUPNAME',
                 'EMAIL',
                 'INITIALS'
                 ]
    
   
    _selectColumns = {2: ['deposition.dep_set_id',
                          'deposition.pdb_id',
                          'deposition.status_code',
                          'deposition.author_release_status_code',
                          'deposition.exp_method',
                          'deposition.annotator_initials',
                          'wf_class_dict.wf_class_id',
                          'wf_class_dict.wf_class_name',
                          'wf_class_dict.version',
                          'wf_instance.wf_inst_id',
                          'wf_instance.owner',
                          'wf_instance.inst_status',
                          'wf_instance.status_timestamp',
                          'wf_task.wf_task_id',
                          'wf_task.task_name',
                          'wf_task.task_status',
                          'wf_task.status_timestamp'
                          ],
                      1: ['DEP_SET_ID',
                          'EXP_METHOD',
                          'PDB_ID',
                          'STATUS_CODE',
                          'AUTHOR_RELEASE_STATUS_CODE',
                          'INITIAL_DEPOSITION_DATE',
                          'STATUS_CODE_EXP',
                          'ANNOTATOR_INITIALS',
                          'AUTHOR_LIST',
                          'SG_CENTER'
                          ]
                     }
    _constraintList={'DEP_SET_ID'          : 'deposition.dep_set_id',
                     'WF_CLASS_ID'         : 'wf_class_dict.wf_class_id',
                     'WF_INST_ID'          : 'wf_instance.wf_inst_id',
                     'WF_TASK_ID'          : 'wf_task.wf_task_id',
                     'STATUS_CODE'         : 'deposition.status_code',
                     'PDB_ID'              : 'deposition.pdb_id',
                     'DEPOSIT_SITE'        : 'deposition.deposit_site',
                     'PROCESS_SITE'        : 'deposition.process_site',
                     'ANNOTATOR_INITIALS'  : 'deposition.annotator_initials',
                     'EXP_METHOD'          : 'deposition.exp_method',
                     'SG_CENTER'           : 'deposition.sg_center',
                     'INST_STATUS'         : 'wf_instance.inst_status',
                     'INST_STATUS_TIMESTP' : 'wf_instance.status_timestamp',
                     'TASK_STATUS_TIMESTP' : 'wf_task.status_timestamp',
                     'TASK_STATUS'         : 'wf_task.task_status',
                     'OWNER'               : 'wf_instance.owner'
                    }
    
    # crossing search from wf_instance,wf_class_dict,deposition and wf_task
    
    _tableJoinSyntext = " FROM wf_instance left join wf_class_dict on " \
                  +"wf_instance.wf_class_id= wf_class_dict.wf_class_id " \
                  +"left join deposition on (wf_instance.wf_class_id=" \
                  +"wf_class_dict.wf_class_id and wf_instance.dep_set_id" \
                  +"=deposition.dep_set_id) left join wf_task on " \
                  +"(wf_task.wf_inst_id=wf_instance.wf_inst_id and " \
                  +"wf_task.dep_set_id=deposition.dep_set_id)"
    
    _orderBy = {1: "order by dep_set_id",
                2: " order by deposition.dep_set_id, wf_instance.wf_inst_id desc," \
                   +"wf_instance.ordinal desc,wf_task.ordinal desc",
                3: ['author_release_status_code','initial_deposition_date']
                }


   
                       
    def __init__(self):
        pass




##
# File:    WfTracking.py
# Date:    27-Apr-2010
#
# Updates:
#  07-Feb-2014   jdw   Move to common path and provide meaningful return code for setInstanceStatus()
#  14-Apr-2014   jdw   Update documentation  -- 
#
##
"""
Classes in this module manage the update of the progress and tracking
database for sequence editing operations.

"""
__docformat__ = "restructuredtext en"
__author__    = "John Westbrook"
__email__     = "jwest@rcsb.rutgers.edu"
__license__   = "Creative Commons Attribution 3.0 Unported"
__version__   = "V0.07"

import os,sys
import datetime

from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi
from wwpdb.utils.wf.dbapi.WFEtime import getTimeNow

class WfTracking(object):
    """Provides methods to update progress and tracking information in the WF status database.

    """
    def __init__(self,verbose=False,log=sys.stderr):
        self.__verbose=verbose
        self.__lfh=log

    def setInstanceStatus(self,depId=None,instId=None,classId=None,status=None):
        """ Update the tracking status record for the input instance step.
        """
        
        now = getTimeNow()
        DBstatusAPI = WfDbApi(verbose=self.__verbose)
        instD = {}
        instD['WF_INST_ID']  = instId
        instD['WF_CLASS_ID'] = classId
        instD['DEP_SET_ID']  = depId
        instD['INST_STATUS'] = status
        instD['STATUS_TIMESTAMP'] = now

        rd = DBstatusAPI.getObject(depId,classId, instId)
        if ((rd is None) or (len(rd) == 0)):
            self.__lfh.write("+WfTracking.setInstanceStatus() no status records for depId %s classId %s instId %s\n" % (depId,classId,instId))
            # new insert --- 
            DBstatusAPI.saveObject(instD, 'insert')
        #
        # This SQL is designed to avoid contention between modules.   No result (0 length) will be returned if 
        #  module is currently running under the current workflow instance id.   If the workflow is running under
        # a different wf instance then we fail here -- 
        #
        #
        sql = "update wf_instance_last set status_timestamp=" + str(now) + ", inst_status='" + str(status) +  "' where dep_set_id = '" + \
              depId + "' and wf_class_id = '" + classId + "' and wf_inst_id = '" + instId + "'"

        ok = DBstatusAPI.runUpdateSQL(sql)
        if ok < 1:
            self.__lfh.write("+WfTracking.setInstanceStatus() ERROR: failed to update workflow status, current task does not control the workflow\n")
            return False
        else:
            # Can update existing record using  ---
            DBstatusAPI.updateStatus(instD,status)
            # Verify the status
            if (self.__verbose):        
                rd = DBstatusAPI.getObject(depId,classId, instId)        
                self.__lfh.write("+WfTracking.setInstanceStatus() verified new status is: %r\n" % DBstatusAPI.getStatus(rd))

        return True

    def setTaskStatus(self,depId=None,instId=None,classId=None,taskId=None,status=None):
        """ Insert a tracking status record for a task step.
        """
        now = getTimeNow()
        taskD = {}
        taskD['WF_TASK_ID']   = taskId
        taskD['WF_INST_ID']   = instId
        taskD['WF_CLASS_ID']  = classId
        taskD['DEP_SET_ID']   = depId
        taskD['TASK_STATUS']  = status
        taskD['STATUS_TIMESTAMP'] = now

        rd = DBstatusAPI.getObject(depId,classId,instId,taskId)
        if ((rd is None) or (len(rd) == 0)):
            self.__lfh.write("+WfTracking.setTaskStatus() no status records for depId %s classId %s instId %s taskId %s\n" %
                             (depId,classId,instId,taskId))
        
        DBstatusAPI.saveObject(taskD, 'insert')

        # To update a current record --- 
        #DBstatusAPI.updateStatus(taskID,"open")
        #DBstatusAPI.updateStatus(taskID,"closed(0)")
        #
        # Verify the status setting
        if (self.__verbose):        
            rd = DBstatusAPI.getObject(depId,classId, instId,taskId)        
            self.__lfh.write("+WfTracking.setTaskStatus() new status is: %r\n" % DBstatusAPI.getStatus(rd))        

        return True

    



"""

    File:    WfDbApi.py

    Providing APIs for workflow engine and workflow manager in D&A tool
 
   __author__    = "Li Chen"
   __email__     = "lchen@rcsb.rutgers.edu"
   __version__   = "V0.01"
   __Date__      = "April 21, 2010"

 Updates: 
  07-Feb-2014  to  add processStatus()
  07-Feb-2014  jdw add socket support
   4-Aug-2014  jdw add debug flag to minimize connection tracking logging
  26-sep-2017  ep  runInsertSQL/runUpdateSQL allow for parameterized arguments
                   to pass to db execute to handle quoting
     
"""

import os,sys
import time
import datetime
import MySQLdb
#
from wwpdb.utils.wf.dbapi.DbConnection import DbConnection    
from wwpdb.utils.wf.dbapi.DbCommand import DbCommand     
from wwpdb.utils.wf.schema.WfSchemaMap  import WfSchemaMap
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.wf.dbapi.WFEtime import getTimeNow

class WfDbApi(object):
    """
      Sample of WfSchemaMap class:
      
       _schemaMap = { 
         "DEPOSITION" : {
                   "ATTRIBUTES"{dictionary of columns}
                   "TABLE_NAME"  : "deposition"
                        }
         "WF_TASK"    : {
                   "ATTRIBUTES"{dictionary of columns}
                   "TABLE_NAME"  : "wf_task"
                        }
                    }
                    
    """
                   
    __schemaWf = WfSchemaMap._schemaMap
    __selectList = WfSchemaMap._selectColumns
    __constraintList = WfSchemaMap._constraintList
    #__statusList = WfSchemaMap._columnForStatus
    __statusList = WfSchemaMap._usefulItems[0:3]
    __columnList = WfSchemaMap._usefulItems
    __tableList = WfSchemaMap._tables
    __idList = WfSchemaMap._objIds
    __refList = WfSchemaMap._referencePairs
    __sqlJoinStr = WfSchemaMap._tableJoinSyntext
    __orderBy = WfSchemaMap._orderBy
    __userInfo = WfSchemaMap._userInfo

    def __init__(self, log=sys.stderr, verbose=False):
        """
          Environmental variable WWPDB_SITE_ID needs to be set in the environment 
          for ConfigInfo() to obtain the correct details -

        """
        
        self.__Nretry = 5
        self.__lfh = log
        self.__verbose = verbose
        self.__debug=False
        cI=ConfigInfo()
        self.__dbServer  = cI.get("SITE_DB_SERVER")
        self.__dbHost    = cI.get("SITE_DB_HOST_NAME")
        self.__dbName    = cI.get("SITE_DB_DATABASE_NAME")
        self.__dbUser    = cI.get("SITE_DB_USER_NAME")
        self.__dbPw      = cI.get("SITE_DB_PASSWORD")
        self.__dbSocket  = cI.get("SITE_DB_SOCKET")
        self.__dbPort    = int("%s" % cI.get("SITE_DB_PORT_NUMBER"))

        if (self.__debug):
            self.__lfh.write("\n+WfDbApi.__init__() using socket %r\n" % self.__dbSocket)
            self.__lfh.write("+WfDbApi.__init__() using socket environment reference %r\n" % os.getenv("SITE_DB_SOCKET",None))


        self.__myDb  = DbConnection(dbServer=self.__dbServer,dbHost=self.__dbHost,
                                    dbName=self.__dbName,
                                    dbUser=self.__dbUser,
                                    dbPw=self.__dbPw,
                                    dbPort=self.__dbPort, 
                                    dbSocket=self.__dbSocket
                                ) 
        
        self.__dbcon = self.__myDb.connect()          
        self.__db = DbCommand(self.__dbcon,self.__lfh,self.__verbose)
        
    
    def reConnect(self):
        """
           Tom : Method to re connect on error with connection
        """
        try:
            self.__myDb.close(self.__dbcon)
        except MySQLdb.Error:
            self.__lfh.write("+WfDbApi.reConnect() DB connection lost - cannot close\n")
            self.__lfh.write("+WfDbApi.reConnect() Re-connecting to the database ..\n")
            self.__lfh.write("+WfDbApi.reConnect() UTC time = %s\n" % datetime.datetime.utcnow())

        for i in range(1,5):
          try:
              self.__dbcon = self.__myDb.connect()
              self.__db = DbCommand(self.__dbcon,self.__lfh,self.__verbose)
              return True
          except MySQLdb.Error:
              self.__lfh.write("+WfDbApi.reConnect() Cannot get re-connection : trying again\n")
              time.sleep(2*i)

        return False
    

    def testConnection(self):
        """
           Test if connection is lost. if yes, rebuild the connection
        """
        
        if(str(self.__dbcon).find("closed")>0):
            
            self.__lfh.write("+WfDbApi::testConnection(): Re-connecting to the database.\n")
            self.__dbcon = self.__myDb.connect()
            self.__db = DbCommand(self.__dbcon,self.__lfh,self.__verbose)
            
    def isConnected(self):
        """  Return boolean flag for connection status -

        """
        try:
            if(str(self.__dbcon).find("open") > 0):
                return True
        except:
            pass
        return False
            
    def close(self):
        """
           This function is only for detecting the connection problem
        """
        
        if(str(self.__dbcon).find("open")>0):
            self.__myDb.close(self.__dbcon)
            if (self.__debug):
                self.__lfh.write("WfDbApi::close(): Closing a connection\n")
            
    def runInsertSQL(self, sql, args = None):
        """
          method to run a query
        """
#        self.testConnection()
        for retry in range(1,self.__Nretry):
            ret = self.__db.runInsertSQL(sql, args)
            if ret == None:
                # database error - 
                if self.__db.dbState > 0:
                    # 2006 : database went away
                    time.sleep(retry*2)  #  backoff for increasing waits
                    if not self.reConnect(): return None
                else:
                    # unhandle DB error
                    return None
            else:
                return ret
        
    def runUpdateSQL(self, sql, args = None):
        """
          method to run a query
        """
#        self.testConnection()
        for retry in range(1,self.__Nretry):
          ret = self.__db.runUpdateSQL(sql, args)
          if ret == None:
# database error - 
            if self.__db.dbState > 0:
# 2006 : database went away
              time.sleep(retry*2)  #  backoff for increasing waits
              if not self.reConnect(): return None
            else:
# unhandle DB error
              return None
          else:
            return ret
 
# all gone bad
        return None

    def runSelectSQL(self,sql):

        """
          method to run a query
        """

#        self.testConnection()

        for retry in range(1,self.__Nretry):
          ret = self.__db.runSelectSQL(sql)
          if ret == None:
# database error - 
            if self.__db.dbState > 0:
# 2006 : database went away
              time.sleep(retry*2)  #  backoff for increasing waits
              if not self.reConnect(): return None
            else:
# unhandle DB error
              return None
          else:
            return ret
 
# all gone bad
          return None
    
    
    def getObject(self, depId=None,classId=None,instId=None,taskId=None):
        """
           Get an object by the depId,,classId,instId,taskId
           return a dictionary of table content
           the table is depended on giving depId,classId,instId and taskId
           
           If all 4 ids are provided, table is wf_task
           if depId,classId,instId provided, but not taskId, table is wf_instance
           if only depId, go to the table deposition
           if classId is provide, depId is empty, the table is wf_class_dict

           Return a dictionary
           The keys are ATTRIBUTES defined in WfSchemaMap._schemaMap and
           The table name is one of (deposition,wf_class_dict,wf_instance,wf_task)
           
        """
        
        tableDef ={}
        rDict={} 
        constraintDict = {}
        orderList=[]
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId= self.checkId(taskId)
        
        if(depId!=None and classId!=None and instId!=None and taskId!=None):
            # table wf_task
            tableDef =self.__schemaWf[self.__tableList[3]]
        elif (depId!=None and classId!=None and instId!=None and taskId==None):
            # table wf_instance
            tableDef =self.__schemaWf[self.__tableList[2]]
        elif (depId==None and classId!=None and instId==None and taskId==None):
            # table wf_class_dict (depId must be None)
            tableDef =self.__schemaWf[self.__tableList[1]]
        elif (depId!=None and classId==None and instId==None and taskId==None):
            # table deposition 
            tableDef =self.__schemaWf[self.__tableList[0]]
        else:
            if self.__verbose:
                self.__lfh.write("WfDbApi::getObject(): Wrong parameters, check all input ids\n")
            return rDict
        
        if(depId!=None):
            constraintDict[self.__idList[0]]  = depId
        if(classId!=None):
            constraintDict[self.__idList[1]]  = classId
        if(instId!=None):
            constraintDict[self.__idList[2]]  = instId
        if(taskId!=None):
            constraintDict[self.__idList[3]]  = taskId
            
        if(tableDef ==self.__schemaWf[self.__tableList[3]] or tableDef ==self.__schemaWf[self.__tableList[2]]):
              orderList.append("ORDINAL")

#        self.testConnection()
        for retry in range(1,self.__Nretry):

          if(len(orderList)>0):
              rDict=self.__db.selectRows(tableDef,constraintDict,orderList)
          else:
              rDict=self.__db.selectRows(tableDef,constraintDict)

          if rDict == None:
# database error - 
            if self.__db.dbState > 0:
# 2006 : database went away
              time.sleep(retry*2)  #  backoff for increasing waits
              if not self.reConnect(): return None
            else:
# unhandle DB error
              return None
          else:
            return rDict

# all gone bad
        return None

    
    def saveObject(self, dataObj, type='insert',constraintDict=None):
        """
           insert/update a new record in the database
           This method currently is for save one of objects in following list
           deposition, class, instance, task             
        """
        
        for i in [3,2,1,0]:
            if(self.__idList[i] in dataObj.keys()):
                dataObj[self.__idList[i]]=self.checkId(dataObj[self.__idList[i]])
                if(dataObj[self.__idList[i]]==None):
                    self.__lfh.write("WfDbApi::saveObject(): %s can not be None or empty string." % self.__idList[i])
                    exit(1)
       
        tableDef = self.getTableDef(dataObj)
        if(len(tableDef)>0):

          for retry in range(1,self.__Nretry):

            ret = self.__db.update(type, tableDef, dataObj, constraintDict)

            if ret == None:
# database error - 
              if self.__db.dbState > 0:
# 2006 : database went away
                time.sleep(retry*2)  #  backoff for increasing waits
                if not self.reConnect(): return None
              else:
# unhandle DB error
                return None
            else:
              return ret

# all gone bad
          return None

        else:
            self.__lfh.write("WfDbApi::saveObject(): The data object is not the one of deposition, class, instance, task\nNothing is done.\n")
            return "bad-code"
        

    def getStatus(self,dataObj):
        """
           Get status from a dataObj(could be a deposition, instance or task obj)
           return a string. if dataObj is not a right type, return an empty str.
        """  
        
        returnObj={}
        oType = str(type(dataObj))
        if(oType.find('dict') > 0 ):
            returnObj=dataObj
        elif(oType.find('list') > 0 ): 
            num = len(dataObj)
            #get the last record
            returnObj=dataObj[num-1]
            
        for k, v in returnObj.items():
            # a task
            if(self.__idList[3] in returnObj.keys() and self.__idList[2] in returnObj.keys() and self.__idList[1] in returnObj.keys() and self.__idList[0] in returnObj.keys()):
                return returnObj[self.__statusList[2]]
            # a instance    
            elif(self.__idList[3] not in returnObj.keys() and self.__idList[2] in returnObj.keys() and self.__idList[1] in returnObj.keys() and self.__idList[0] in returnObj.keys()):
                return returnObj[self.__statusList[1]]
            # a deposition    
            elif(self.__idList[0] in returnObj.keys() and self.__idList[1] not in returnObj.keys() and self.__idList[2] not in returnObj.keys() and self.__idList[3] not in returnObj.keys()):
                return returnObj[self.__statusList[0]]
                
            else:
                pass
        
        return  ""


    
    def updateStatus(self, dataObj, status=None):
        """
           Change status for one entry in table deposition, wf_instance or wf_task
           
        """
        type = 'update'
        if(status ==None or status ==""):
            self.__lfh.write("WfDbApi::updateStatus(): Failing - no status code given\n")
            exit (1)
            
        tableDef = self.getTableDef(dataObj)
        if(len(tableDef)>0):
            updateVal= {}
            updateVal = self.assignStatus(tableDef, status)
            
            rDict={}
            constraintDict = {}
            if(self.__idList[0] in dataObj.keys() and dataObj[self.__idList[0]]!=None and dataObj[self.__idList[0]]!=""):
                constraintDict[self.__idList[0]]  = dataObj[self.__idList[0]]
            if(self.__idList[1] in dataObj.keys() and dataObj[self.__idList[1]]!=None and dataObj[self.__idList[1]]!=""):
                constraintDict[self.__idList[1]]  = dataObj[self.__idList[1]]
            if(self.__idList[2] in dataObj.keys() and dataObj[self.__idList[2]]!=None and dataObj[self.__idList[2]]!=""):
                constraintDict[self.__idList[2]]  = dataObj[self.__idList[2]]
            if(self.__idList[3] in dataObj.keys() and dataObj[self.__idList[3]]!=None and dataObj[self.__idList[3]]!=""):
                constraintDict[self.__idList[3]]  = dataObj[self.__idList[3]]
                
#            self.testConnection()
            for retry in range(1,self.__Nretry):

              rDict=self.__db.update(type, tableDef, updateVal, constraintDict)

              if rDict == None:
# database error - 
                if self.__db.dbState > 0:
# 2006 : database went away
                  time.sleep(retry*2)  #  backoff for increasing waits
                  if not self.reConnect(): return None
                else:
# unhandle DB error
                  return None
              else:
                return "ok"

# all gone bad
            return None

        else:
            self.__lfh.write("+WfDbApi::updateStatus(): The data object is not the one of deposition, instance, task. Nothing is updated.\n")
            return "code-bad"
            
    def processStatus(self,depID,instID,classID):

      '''
        status owner of process 
        None : this depID/instID/classID does NOT own the current process
        <value> : the status 
      '''

      try:
        sql = "select inst_status from dep_last_instance where dep_set_id = '" + str(depID) + "' AND inst_id = '" + str(instID) + "' AND class_id = '" + str(classID) + "'";
        print(sql)
        rows = self.runSelectSQL(sql);
        self.close()
        for row in rows:
            return row[0]

        return None
      except Exception as e:
        print("Exception in processOwner " + str(e))
        return None
        
    def getReference(self,depId=None,classId=None,instId=None,taskId=None):
        """
           Get a list of reference data from table wf_reference
           An empty id will be treated as None.

           Retrun a rwo list or dictionary (if only one record)
        """
        
        tableDef = self.__schemaWf[self.__tableList[4]]
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId= self.checkId(taskId)
        constraintDict = {}
        if(depId!=None):
            constraintDict[self.__idList[0]]  = depId
        else:
            constraintDict[self.__idList[0]] = 'None'
        if(classId!=None):
            constraintDict[self.__idList[1]]  = classId
        else:
            constraintDict[self.__idList[1]] = 'None'
        if(instId!=None):
            constraintDict[self.__idList[2]]  = instId
        else:
            constraintDict[self.__idList[2]] = 'None'
        if(taskId!=None):
            constraintDict[self.__idList[3]]  = taskId
        else:
            constraintDict[self.__idList[3]] = 'None'

#        self.testConnection()
        for retry in range(1,self.__Nretry):

          results=self.__db.selectRows(tableDef,constraintDict)

          if results == None:
# database error - 
            if self.__db.dbState > 0:
# 2006 : database went away
              time.sleep(retry*2)  #  backoff for increasing waits
              if not self.reConnect(): return None
            else:
# unhandle DB error
              return None
          else:
            return results

# all gone bad
        return None

    
    
    def addReference(self,type,depId=None,classId=None,instId=None,taskId=None,hashId=None,hashVal=None):
        """
           Add or update reference record in the table wf_reference

           requirement:
           For a deposition level reference,the classId, instId and taskId should be None
           For a class level reference the instId, taskId should be None
           For a instance level reference taskId should be None
           For a task level reference all ids ahould be Non-null.
           
           
        """
        
        tableDef = self.__schemaWf[self.__tableList[4]]
        constraintDict = {}
        updateVal={}
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId= self.checkId(taskId)

        type = type.lower()
        if(type=='update'):
            if(depId!=None):
                constraintDict[self.__idList[0]]  = depId
            else:
                constraintDict[self.__idList[0]] = 'None'
            if(classId!=None):
                constraintDict[self.__idList[1]]  = classId
            else:
                constraintDict[self.__idList[1]] = 'None'
            if(instId!=None):
                constraintDict[self.__idList[2]]  = instId
            else:
                constraintDict[self.__idList[2]] = 'None'
            if(taskId!=None):
                constraintDict[self.__idList[3]]  = taskId
            else:
                constraintDict[self.__idList[3]] = 'None'
            if(hashId!=None and self.__refList[0] in tableDef['ATTRIBUTES'].keys()):
                constraintDict[self.__refList[0]]=hashId    
        else:
            if(depId!=None):
                updateVal[self.__idList[0]]  = depId
            if(classId!=None):
                updateVal[self.__idList[1]]  = classId
            if(instId!=None):
                updateVal[self.__idList[2]]  = instId
            if(taskId!=None):
                updateVal[self.__idList[3]]  = taskId
                
            if(hashId!=None and self.__refList[0] in tableDef['ATTRIBUTES'].keys()):
                updateVal[self.__refList[0]]=hashId
                
        if(hashVal!=None and self.__refList[1] in tableDef['ATTRIBUTES'].keys()):
            updateVal[self.__refList[1]]=hashVal
            
#        self.testConnection()    
        for retry in range(1,self.__Nretry):

          ret = self.__db.update(type,tableDef,updateVal,constraintDict)

          if ret == None:
# database error - 
            if self.__db.dbState > 0:
# 2006 : database went away
              time.sleep(retry*2)  #  backoff for increasing waits
              if not self.reConnect(): return None
            else:
# unhandle DB error
              return None
          else:
            return "ok"

# all bad
        return None

    def checkId(self, str):
        """
           Test if an object id is empty.
           If yes return None, else return str.
        """
        if(str ==""):
            return None
        else:    
           return str
    
    
    def exist(self, dataObj):
        """
           Test if a dataObj exists in the Database based on IDs
           dataObj is only for one of objects(deposition, class, instance, task).
           Return Ture or False
        """

        existObj=False
        depId=None
        classId=None
        instId=None
        taskId=None
        if(len(dataObj)==0):
            existObj=False 
        else:
            if(self.__idList[3] in dataObj.keys()):
                taskId = dataObj[self.__idList[3]]
            if(self.__idList[2] in dataObj.keys()):
                instId = dataObj[self.__idList[2]]
            if(self.__idList[1] in dataObj.keys()):
                classId = dataObj[self.__idList[1]]
            if(self.__idList[0] in dataObj.keys()):
                depId = dataObj[self.__idList[0]]
            if(depId==None and classId==None and instId==None and taskId==None):
                existObj = False
            else:    
                rd = self.getObject(depId,classId,instId,taskId)
                
            if(len(rd) >0):
                existObj = True
            else:
                existObj = False
                

        return existObj
            
         
    def getTableDef(self, dataObj):
        """
           Determine which table to use in the database

           If all 4 ids are in the dataObj, table is wf_task
           if instId is provided, but not taskId, table is wf_instance
           if only depId, go to the table deposition
           if classId is provide, depId is empty, the table is wf_class_dict
           Return table dict in __schemaWf
        """

        tableDef={}
        if(self.__idList[3] in dataObj.keys() and self.__idList[2] in dataObj.keys() and \
           self.__idList[1] in dataObj.keys() and self.__idList[0] in dataObj.keys()):
            tableDef = self.__schemaWf[self.__tableList[3]]
        if(self.__idList[3] not in dataObj.keys() and self.__idList[2] in dataObj.keys() and\
           self.__idList[1] in dataObj.keys() and self.__idList[0] in dataObj.keys()):
            tableDef = self.__schemaWf[self.__tableList[2]]
        if(self.__idList[3] not in dataObj.keys() and self.__idList[2] not in dataObj.keys()\
           and self.__idList[1] in dataObj.keys() and self.__idList[0] not in dataObj.keys()):
            tableDef = self.__schemaWf[self.__tableList[1]]
        if(self.__idList[3] not in dataObj.keys() and self.__idList[2] not in dataObj.keys()\
           and self.__idList[1] not in dataObj.keys() and self.__idList[0] in dataObj.keys()):
            tableDef = self.__schemaWf[self.__tableList[0]]

        return tableDef
    
    
    def assignStatus(self, tableDef, status):
        """
           Set status value for table in tableDef
           
           Return: a dictionary of update Values
           
        """
        
        attribDict =  tableDef['ATTRIBUTES']
        updateDict={}
        for k in range(len(self.__statusList)):
            if(self.__statusList[k] in attribDict.keys()):
                updateDict[self.__statusList[k]] = status
# Tom : added change to timestamp when we make a status change
# Tom : only valid for task and instance : check schema.usefulItems ending in status
                if self.__statusList[k].endswith('STATUS'):
                  updateDict['STATUS_TIMESTAMP'] = getTimeNow()

        return  updateDict
    
    
    def makeDataDict(self, tableDef, dataObj):
        """
           Make a dictionary to hold update status using table columns as keys
           
           Return: a dictionary of update values
           
        """
        
        attribDict =  tableDef['ATTRIBUTES']
        dataDict={}
        objType = str(type(dataObj))

        if (objType.find('dict') > 0 ):
            for k,v in dataObj.items():
                if(k in attribDict.keys()):
                    dataDict[k] = v
        else:
            self.__lfh.write("WfDbApi::makeDataDict(): Failing, input data object is not a dictionary.\n")

        return dataDict

    def getAll(self,depId=None,classId=None,instId=None):
        """
           Get info from WF table deposition, wf_class_dict, wf_instance.

           Return a list of rows (python dictionaries) 
           The Keys in the result dictionary is defined in
           WfSchemaMap._selectColumns, could be in the following:
               deposition.dep_set_id
               deposition.pdb_id
               deposition.status_code
               deposition.author_release_status_code
               deposition.exp_method
               deposition.annotator_initials
               wf_class_dict.wf_class_id
               wf_class_dict.wf_class_name
               wf_class_dict.version
               wf_instance.wf_inst_id
               wf_instance.owner
               wf_instance.inst_status
               wf_instance.status_timestamp
               wf_task.wf_task_id
               wf_task.task_name
               wf_task.task_status
               wf_task.status_timestamp
        """
        
        constraintDef={}
        """
           Assumption is that you can't choose classId or instId without
           depId. Also any instId must have depId and classId.
        """
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
    
        if(depId==None and (classId!=None or instId!=None)):
            self.__lfh.write("+WfDbApi::getAll(): Failing, no deposition id provided\n")
            exit(1)
        if(depId!=None and classId==None and instId!=None):
            self.__lfh.write("+WfDbApi::getAll(): Failing, no class id provided\n")
            exit(1)
           
        if(depId!=None):
            constraintDef[self.__idList[0]]=depId
        if(classId!=None):
            constraintDef[self.__idList[1]]=classId
        if(instId!=None):
            constraintDef[self.__idList[2]]=instId
            
#        self.testConnection()
        for retry in range(1,self.__Nretry):

            results=self.__db.selectCrossTables(self.__selectList[2],self.__sqlJoinStr,self.__orderBy[2], self.__constraintList,constraintDef)

            if results == None:
# database error - 
              if self.__db.dbState > 0:
# 2006 : database went away
                time.sleep(retry*2)  #  backoff for increasing waits
                if not self.reConnect(): return None
              else:
# unhandle DB error
                return None
            else:
              return results

# all gone bad
        return None


    def doQuery(self,level, parameterDict, orderList=[], otherOpt=None):
        """
            Function for determining which table to query based on level.
            
            Level 1 id for summary information; level 2 is for instance/task
            information.

            parameterDict{} holds parameters that user select from the front end.
            The keys should match the keys in WfSchemaMap._constraintList.

            orderList should also match the keys _schemaMap._orderBy
            or WfSchemaMap._constraintList{} for level 2 
            
            otherOpt could be one of ['AUTHOR_CORRECTIONS','DEP_WITH_PROBLEMS',
            'RELEASE_REQUEST'].
            
            Return a list of rows (dictionaries)
            The Keys in the result dictionary could be in the following:
            Level 1:
               DEP_SET_ID
               EXP_METHOD
               PDB_ID
               STATUS_CODE
               AUTHOR_RELEASE_STATUS_CODE
               INITIAL_DEPOSITION_DATE
               STATUS_CODE_EXP
               ANNOTATOR_INITIALS
               AUTHOR_LIST
               SG_CENTER
               ASSESSION_CODE
               RELATIONSHIP
               ASSOCIATED_IDS
               CORRECTIONS ( for Author's Corrections)
               REQ_CITATION ( for entries requested for release)
               PROBLEM_TYPE ( for Peorblem/Error entries)
               PROBLEM_DETAILS ( for Peorblem/Error entries)
            Level 2: 
               deposition.dep_set_id
               deposition.pdb_id
               deposition.status_code
               deposition.author_release_status_code
               deposition.exp_method
               deposition.annotator_initials
               wf_class_dict.wf_class_id
               wf_class_dict.wf_class_name
               wf_class_dict.version
               wf_instance.wf_inst_id
               wf_instance.owner
               wf_instance.inst_status
               wf_instance.status_timestamp
               wf_task.wf_task_id
               wf_task.task_name
               wf_task.task_status
               wf_task.status_timestamp
        """

        tableDef=""
        rList=[]
        if(level==2):
            # Multiple tables selected
            # The selectList and query are fixed
            if (otherOpt != None):
                self.__lfh.write("+WfDbApi::doQuery(): Failing, otherOpt is only for level 1\n")
                exit(1)
            for k,v in parameterDict.items():
                if(k not in self.__constraintList.keys()):
                    self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % (k))
                    exit(1)
            if(len(orderList)>0):
                for k in orderList:
                    if(k not in self.__constraintList.keys()):
                        self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % (k))
                        exit(1)
                orderBy=self.__db.makeOrderStr(orderList)
            else:
                #use default
                orderBy=self.__orderBy[2]

#            self.testConnection()    
            ok = False
            for retry in range(1,self.__Nretry):

               rList=self.__db.selectCrossTables(self.__selectList[2],self.__sqlJoinStr,\
                                               orderBy,self.__constraintList,parameterDict)

               if rList == None:
# database error - 
                 if self.__db.dbState > 0:
# 2006 : database went away
                   time.sleep(retry*2)  #  backoff for increasing waits
                   if not self.reConnect(): return None
                 else:
# unhandle DB error
                   return None
               else:
                 ok = True
                 break;

# all gone bad
            if not ok: return None
        else:
            # level 1, query single tables, 
            # first get data from the table "DEPOSITION" 
            tableDef=self.__schemaWf[self.__tableList[0]]
            for k in parameterDict.keys():
                if(k not in tableDef['ATTRIBUTES']):
                    self.__lfh.write("+WfDbApi::doQuery(): Failing, no matched columns in the database for %s\n" % (k))
                    exit(1)
            if(otherOpt != None):
                if(otherOpt in (self.__tableList[9],self.__tableList[10],self.__tableList[11])):
                    sqlStr = self.__idList[0]+" in (select "+self.__idList[0]+" from " +self.__schemaWf[otherOpt]['TABLE_NAME']+ ")"
                    parameterDict['EXTERNAL_TABLE']= sqlStr
                    #  EXTERNAL_TABLE is for special SQL syntext
                else:
                    self.__lfh.write("+WfDbApi::doQuery(): Failing, otherOpt should be in the list [$s,%s,%]s\n" % (self.__tableList[9],self.__tableList[10],self.__tableList[11]))
                    exit(1)
                    
            orderList = self.__orderBy[3]
#            self.testConnection()

            ok = False
            for retry in range(1,self.__Nretry):

               results=self.__db.selectRows(tableDef,parameterDict,orderList,self.__selectList[1])

               if results == None:
# database error - 
                 if self.__db.dbState > 0:
# 2006 : database went away
                   time.sleep(retry*2)  #  backoff for increasing waits
                   if not self.reConnect(): return None
                 else:
# unhandle DB error
                   return None
               else:
                 ok = True
                 break;

# all gone bad
            if not ok: return None
                
            # convert dict to list if there is only one record.
            if(str(type(results)).find('dict') > 0):
                rList.append(results)
            else:
                rList=results

            # get info other then the main table
                
            for k in rList:
                Relations=""
                associatedIds=""
                constDict ={}
                DepId = k[self.__selectList[1][0]]
                DepId = self.checkId(DepId)
                if(DepId != None):
                    constDict[self.__idList[0]] = DepId
                    # get 'ASSESSION_CODE' from the table "DATABASE_REF"
                    accessionIds =  self.getValueString(self.__schemaWf[self.__tableList[7]],constDict,self.__columnList[3])
                    if(accessionIds != None and accessionIds !=""):
                        # fill in 'ASSESSION_CODE' in the result
                        k[self.__columnList[11]]=accessionIds

                    # get 'REPLACE_PDB_ID' from table "DATABASE_PDB_OBS_SPR" 
                    obsIds = self.getValueString(self.__schemaWf[self.__tableList[6]],constDict,self.__columnList[4])
                    if(obsIds != None and obsIds !=""):
                        Relations +="SPR/OBS"
                        associatedIds += obsIds
                    
                    # get related 'DB_ID' from table "DATABASE_RELATED"
                    # only "SPLIT" is interested
                    constDict[self.__columnList[6]] = "SPLIT"
                    relatedIds = self.getValueString(self.__schemaWf[self.__tableList[8]],constDict,self.__columnList[7])
                    
                    if(relatedIds != None and relatedIds !=""):
                        if(Relations !=""):
                            Relations +=", SPLIT"
                            associatedIds += ", "+relatedIds
                        else:
                            Relations +="SPLIT"
                            associatedIds += relatedIds
                            
                    if(Relations != None and Relations != ""):
                        # fill in RELATIONSHIP and ASSOCIATED_IDS in the result
                        k[self.__columnList[9]]=Relations
                        k[self.__columnList[10]]=associatedIds

                    if(otherOpt != None):
                        constDict={}
                        constDict[self.__idList[0]] = DepId
                        selectItem =""
                        if(otherOpt == self.__tableList[9]):
                            # AUTHOR_CORRECTIONS.CORRECTIONS
                            selectItem = self.__columnList[12]
                        if(otherOpt == self.__tableList[10]):
                            # RELEASE_REQUEST.CITATION
                            selectItem = self.__columnList[13]
                        if(otherOpt == self.__tableList[11]):
                            # DEP_WITH_PROBLEMS.PROBLEM_TYPE
                            selectItem = self.__columnList[14]
                            
                        resultString =  self.getValueString(self.__schemaWf[otherOpt],constDict,selectItem)
                        if(resultString != None and resultString !=""):
                        # fill selectItem in the result
                            k[selectItem]=resultString
                            
                        if(otherOpt == self.__tableList[11]):
                            # one more items: DEP_WITH_PROBLEMS.PROBLEM_DETAILS
                            selectItem = self.__columnList[15]
                            resultString =  self.getValueString(self.__schemaWf[otherOpt],constDict,selectItem)
                            if(resultString != None and resultString !=""):
                                k[selectItem]=resultString
                            
        return  rList
    
    
    def getValueString(self, tableDef,constDict,selectItem):
        """
           Query a single table for a giving DEP_SET_ID
           selectItem is the selected column for query

           return a string (if there are multiple results
           the string is concatenated by ", "
        """

        returnString=""
        
        # define the check list 'DATABASE_CODE','REPLACE_PDB_ID','DB_ID'
        # 'CORRECTIONS','REQ_CITATION','PROBLEM_TYPE','PROBLEM_DETAILS',
        # 'PUBMED_ID'
        checkList = [self.__columnList[3], self.__columnList[4],\
                     self.__columnList[7], self.__columnList[12], \
                     self.__columnList[13], self.__columnList[14], \
                     self.__columnList[15],self.__columnList[16]]
        selectList=[]
        if(selectItem !=None and selectItem in checkList):
            selectList.append(selectItem)
        else:
            #print "WfDbApi::getValueString(): Warning -- There is not selectItem"
            pass

#        self.testConnection()
        ok = False
        for retry in range(1,self.__Nretry):

            results=self.__db.selectRows(tableDef,constDict,[],selectList)

            if results == None:
# database error - 
              if self.__db.dbState > 0:
# 2006 : database went away
                time.sleep(retry*2)  #  backoff for increasing waits
                if not self.reConnect(): return None
              else:
# unhandle DB error
                return None
            else:
              ok = True
              break;

        if not ok: return None

        if(str(type(results)).find('dict') > 0 ):
            for checkItem in checkList:
                if( checkItem in results.keys()):
                    returnString=results[checkItem]
        else:
            for k in results:
                for checkItem in checkList:
                    if(checkItem in k.keys()):
                        if(returnString ==""):
                            returnString =k[checkItem]
                        else:
                            returnString += ", "+k[checkItem]
                            
        return returnString
    

    def getNextWfInstId(self, depId, classId):
        """
           get a WF instance ID in highest number

           return a integer (instId+1)
        """

        rDict={}
        returnId=0
        rDict=self.getLastObjectOfState(None,depId,classId)
        if(len(rDict) == 0):
            # no instance
            return 1
        else:
            returnId = rDict[self.__idList[2]]
            return int("%s" % returnId[2:])+1
            

    def referenceExist(self,depId=None,classId=None,instId=None,taskId=None,hashId=None,hashVal=None):

        """
           Check to see if the reference data still exists

        """

        tableDef = self.__schemaWf[self.__tableList[4]]
        constraintDict = {}
        selectList = {}
        orderList = {}
        depId = self.checkId(depId)
        classId = self.checkId(classId)
        instId = self.checkId(instId)
        taskId= self.checkId(taskId)

        if(depId!=None):
            constraintDict[self.__idList[0]]  = depId
        else:
            constraintDict[self.__idList[0]] = 'None'
        if(classId!=None):
            constraintDict[self.__idList[1]]  = classId
        else:
            constraintDict[self.__idList[1]] = 'None'
        if(instId!=None):
            constraintDict[self.__idList[2]]  = instId
        else:
            constraintDict[self.__idList[2]] = 'None'
        if(taskId!=None):
            constraintDict[self.__idList[3]]  = taskId
        else:
            constraintDict[self.__idList[3]] = 'None'    

        if(hashId!=None and self.__refList[0] in tableDef['ATTRIBUTES'].keys()):
            constraintDict[self.__refList[0]]=hashId
        if(hashVal!=None and self.__refList[1] in tableDef['ATTRIBUTES'].keys()):
            selectList[self.__refList[1]]=hashVal

#        self.testConnection()
        for retry in range(1,self.__Nretry):

            results = self.__db.selectRows(tableDef,constraintDict, orderList, selectList)

            if results == None:
# database error - 
              if self.__db.dbState > 0:
# 2006 : database went away
                time.sleep(retry*2)  #  backoff for increasing waits
                if not self.reConnect(): return None
              else:
# unhandle DB error
                return None
            else:
              if(len(results) > 0):
                  return True
              else:
                  return False

# all gone bad
        return None


    def addReferenceOverwrite(self,depId=None,classId=None,instId=None,taskId=None,hashId=None, hashVal=None):
        """
           This method will insert if the data is new, or update if the data exists

           Add or update reference record in the table wf_reference
           requirement:
           For a deposition level reference,the classId, instId and taskId should be None
           For a class level reference the instId, taskId should be None
           For a instance level reference taskId should be None
           For a task level reference all ids ahould be Non-null.
        """
        
        refExist = self.referenceExist(depId,classId,instId,taskId,hashId,hashVal)

        if refExist:
            self.addReference("update",depId,classId,instId,taskId,hashId,hashVal)
        else:
            self.addReference("insert",depId,classId,instId,taskId,hashId,hashVal)            


    def getLastObjectOfState(self,state,depId,classId,instId=None):
        """
           Get the last instance or task by the depId,,classId and instId
           the table is depended on giving depId,classId,instId
           
           If instId=None get the last instance from wf_instance; 
           if instId != None table is wf_task

           State could be any status in wf_instance or wf_task or None
           Return a dictionary of table content
           
        """
        
        tableDef ={}
        rDict={} 
        constraintDict = {}
        orderList=[]
        depId = self.checkId(depId)
        if(depId==None):
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- No depId provided\n")
            sys.exit (1)
        classId = self.checkId(classId)
        if(classId==None):
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- No classId provided\n")
            sys.exit (1)
        if(instId!=None):
            instId = self.checkId(instId)
            
        if(depId!=None and classId!=None and instId!=None):
            # table wf_task
            tableDef =self.__schemaWf[self.__tableList[3]]
            
        elif(depId!=None and classId!=None and instId==None):
            # table wf_instance
            tableDef =self.__schemaWf[self.__tableList[2]]
        else:
            self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- Can not decide the table name\n")
            return rDict
       
        if(depId!=None):
            constraintDict[self.__idList[0]]  = depId
        if(classId!=None):
            constraintDict[self.__idList[1]]  = classId
        if(instId!=None):
            constraintDict[self.__idList[2]]  = instId
        if(state!=None):
            if(tableDef ==self.__schemaWf[self.__tableList[3]]):
                #wf_task
                constraintDict[self.__statusList[2]] = state
               
            elif(tableDef ==self.__schemaWf[self.__tableList[2]]):  
                #wf_instance
                constraintDict[self.__statusList[1]] = state
            else:
                self.__lfh.write("+WfDbApi::getLastObjectOfState(): WARNING -- Can not decide the column name about state\n")
                return rDict
      
        orderList.append("ordinal desc")
        
#        self.testConnection()
        for retry in range(1,self.__Nretry):

            rDict=self.__db.selectRows(tableDef,constraintDict,orderList)

            if rDict == None:
# database error - 
              if self.__db.dbState > 0:
# 2006 : database went away
                time.sleep(retry*2)  #  backoff for increasing waits
                if not self.reConnect(): return None
              else:
# unhandle DB error
                return None
            else:
              if((str(type(rDict)).find('list')) > 0):
                  return rDict[0]
              else:
                  return rDict

# all gone bad
        return None


if __name__ == '__main__':
    pass



##
# File:  dbApi.py
# Date:  14-Feb-2015
#
# #

import sys
from wwpdb.utils.wf.dbapi.WfDbApi import WfDbApi

'''

Grab bag of methods to execute SQL commands on various WF status tables -

'''


class dbAPI(object):

    def __init__(self, depID, connection=None, verbose=False):

        if connection:
            self.con = connection
        else:
            self.con = WfDbApi(verbose=False)
        self.depID = depID
        self.verbose = verbose

    def close(self):

        self.con.close()

    def runSelectNQ(self, table=None, join=None, select=None, where=None, order=None, reverse=False, ordinal=0, run=True, limit=0):

        if where:
            for k, v in where.iteritems():
                where[k] = "'" + v + "'"
        return self.runSelect(table=table, join=join, select=select, where=where, order=order, reverse=reverse, ordinal=ordinal, run=run, limit=limit)

    def runSelect(self, table=None, join=None, select=None, where=None, order=None, reverse=False, ordinal=0, run=True, limit=0):
        '''
          Very simple select SQL creator - single table
          table = string
          values = list
          where = dictionary of equivalence
          order = list

          Joins are specified with multiple table string , and where clause with inner join defined

          IMPT : all attribute values MUST BE QUOTED if strings

          Returns a list of list  (rows of data columns)
          Returns an empty list if nothing returned
        '''

        if not table:
            print "WFE.dbAPI.runSelect : Undefined table"
            return []

        if not select:
            print "WFE.dbAPI.runSelect : Undefined select"
            return []

        depDB = {}
        depDB["DEP_SET_ID"] = self.depID

        try:
            if self.con.exist(depDB):
                #       if True:
                sql = "select " + ','.join(select) + ' from ' + str(table) + ' '
                if where:
                    sql += ' where ' + ' and '.join(['%s = %s' % (k, v) for k, v in where.iteritems()])
                if join:
                    sql += ' and ' + join
                if ordinal > 0:
                    sql += ' and ordinal >= ' + str(ordinal)
                if order:
                    sql += ' order by ' + ','.join(order)
                    if reverse:
                        sql += ' desc '
                if limit > 0:
                    sql += ' limit ' + str(limit)
                if self.verbose:
                    print "WFE.dbAPI.runSelect > " + str(sql)
                if run:
                    ret = self.con.runSelectSQL(sql)
                    return ret
                else:
                    return sql
            else:
                return []
        except Exception as e:
            print "WFE.dbAPI.runSelect :Exception " + str(e)
            return []

    def runUpdateOnOrdinal(self, table=None, ordinal=None, data=None, run=True):
        '''
          Very simple update/insert SQL creator base on ordinal
          table = string
          data = values to insert/update
          where clause is hard wired to ordinal (unique) so can only UPDATE
              a bad ordinal returns 0 rows udpated

          return number of rows updated
          if run = False : returns the SQL

          IMPT : all attribute values MUST BE QUOTED if strings
        '''

        if not table:
            print "WFE.dbAPI.runUpdateOnOrdinal : Undefined table"
            return False

        if not ordinal:
            print "WFE.dbAPI.runUpdateOnOrdinal : Undefined ordinal"
            return False

        try:
            sql = "update " + str(table) + " set " + ','.join(['%s = %s' % (k, v) for k, v in data.iteritems()])
            sql = sql + " where ordinal = " + str(ordinal)
            if self.verbose:
                print "WFE.dbAPI.runInsertUpdate(update) > " + str(sql)

            if run:
                ok = self.con.runUpdateSQL(sql)
                if not ok:
                    print "WFE.dbAPI.runSelect :False to update/insert data " + str(sql)
                return ok
            else:
                return sql
        except Exception as e:
            print "WFE.dbAPI.runSelect :Exception " + str(e)
            return False

    def runInsertUpdateNQ(self, table=None, depID=None, where=None, data=None, run=True):
        '''
          wrapper to add quotes to values
        '''

        if where:
            for k, v in where.iteritems():
                where[k] = "'" + v + "'"
        if data:
            for k, v in data.iteritems():
                data[k] = "'" + v + "'"
        return self.runInsertUpdate(table, depID, where, data, run)

    def runInsertNQ(self, table=None, depID=None, where=None, data=None, run=True):

        if where:
            for k, v in where.iteritems():
                where[k] = "'" + v + "'"
        if data:
            for k, v in data.iteritems():
                data[k] = "'" + v + "'"
        return self.runInsert(table, depID, where, data, run)

    def runInsert(self, table=None, depID=None, where=None, data=None, run=True):

        try:
            if depID:
                sql = "insert into " + str(table) + " (dep_set_id," + ','.join(['%s' % (k) for k, v in data.iteritems()])
                sql += ") values ('" + str(depID) + "'," + ','.join(['%s' % (v) for k, v in data.iteritems()]) + ")"

            if where:
                sql = "insert into " + str(table) + " (" + ','.join(['%s' % (k) for k, v in where.iteritems()])
                if data:
                    sql += "," + ','.join(['%s' % (k) for k, v in data.iteritems()])
                sql += ") values (" + ','.join(['%s' % (v) for k, v in where.iteritems()])
                if data:
                    sql += "," + ','.join(['%s' % (v) for k, v in data.iteritems()])
                sql += ")"

            if self.verbose:
                print "WFE.dbAPI.runInsertUpdate(insert) > " + str(sql)

            if run:
                return self.con.runInsertSQL(sql)
            else:
                return sql
        except Exception as e:
            print "WFE.dbAPI.runInsert :Exception " + str(e)
            return False

    def runUpdate(self, table=None, depID=None, where=None, data=None, run=True):

        try:
            sql = "update " + str(table) + " set " + ','.join(['%s = %s' % (k, v) for k, v in data.iteritems()])
            if depID:
                sql += " where dep_set_id = '" + str(depID) + "'"
            if where:
                sql += ' where ' + ' and '.join(['%s = %s' % (k, v) for k, v in where.iteritems()])

            if self.verbose:
                print "WFE.dbAPI.runInsertUpdate(update) > " + str(sql)

            if run:
                return self.con.runUpdateSQL(sql)
            else:
                return sql

        except Exception as e:
            print "WFE.dbAPI.runUpdate :Exception " + str(e)
            return False

    def runInsertUpdate(self, table=None, depID=None, where=None, data=None, run=True):
        '''
          Very simple update/insert SQL creator
          table = string
          data = values to insert/update
          where clause is hard wired to dep_set_id so can ONLy be used for
              for tables with unique rows on dep_set_id
              Ie if the table contains a row with depsetid - then it is update
                    if not - the data is inserted

          return number of rows updated
          if run = False : returns the SQL

          if depID is not none - then we test the existance based on depID and the table MUST BE UNIQUE on depID
            data is added to the row based on the dep_set_id + data
          if unique is not none - then we test the existance on the synthetic key based on the unique dictionary
            data is added to the row based on the where + data


          IMPT : all attribute values MUST BE QUOTED if strings
        '''

        if not table:
            print "WFE.dbAPI.runUpdateOnOrdinal : Undefined table"
            return False

        rowExists = False
        if depID:
            # then data is unique on depID
            depDB = {}
            depDB["DEP_SET_ID"] = depID
            if self.con.exist(depDB):
                rowExists = True

        else:
            if where:
                sql = "select ordinal from " + str(table) + " where " + ' and '.join(['%s = %s' % (k, v) for k, v in where.iteritems()])
                rows = self.con.runSelectSQL(sql)
                if rows and len(rows) > 0:
                    rowExists = True
            else:
                print "WFE.dbAPI.runUpdate: Undefined key "

        ok = True
        try:
            if rowExists:
                ok = self.runUpdate(table, depID, where, data, run)

            else:
                ok = self.runInsert(table, depID, where, data, run)

            if not run:
                return ok

            if not ok:
                print "WFE.dbAPI.runSelect :False to update/insert data "
                return ok
            else:
                return ok
        except Exception as e:
            print "WFE.dbAPI.runSelect :Exception " + str(e)
            return False


def main(argv):

    print "starting DBAPI test"

    depid = 'D_1100201819'
    ss = dbAPI(depid, verbose=True)

# table join test - notice that in the where clause be careful with the key must not repeat - so be carful with the inner join
    ret = ss.runSelect(
        table='deposition,wf_instance',
        select=[
            'deposition.dep_set_id',
            'wf_instance.wf_inst_id'],
        where={
            "deposition.dep_set_id": "wf_instance.dep_set_id",
            "wf_instance.dep_set_id": "'" +
            depid +
            "'",
            "wf_instance.wf_inst_id": "'W_001'"},
        run=True)
    print str(ret)

    ret = ss.runSelectNQ(table='deposition', select=['ordinal', 'dep_set_id', 'depPW', 'annotator_initials'], where={"dep_set_id": depid})
    print str(ret)


# test the orinal updates
    ret = ss.runSelectNQ(table='deposition', select=['ordinal', 'dep_set_id', 'depPW'], where={"dep_set_id": depid})
    print str(ret)
#   ret = ss.runUpdateOnOrdinal(table='deposition',ordinal='17823',data={"depPW":"'abcdef'"},run=True)
#   print str(ret)
    ret = ss.runSelectNQ(table='deposition', select=['ordinal', 'dep_set_id', 'depPW'], where={"dep_set_id": depid})
    print str(ret)
#   ret = ss.runUpdateOnOrdinal(table='deposition',ordinal='17823',data={"depPW":"'123456'"},run=True)
    print str(ret)
    ret = ss.runSelectNQ(table='deposition', select=['ordinal', 'dep_set_id', 'depPW'], where={"dep_set_id": depid})
    print str(ret)

    ret = ss.runSelectNQ(
        table='deposition,user_data',
        select=[
            'last_name',
            'role',
            'user_data.country'],
        join='deposition.dep_set_id=user_data.dep_set_id',
        where={
            "deposition.dep_set_id": depid})

# test the dep_set_id updates
#   ret = ss.runInsertUpdate(table='deposition',depID='D_1000200025',data={"depPW":"'abcdef'"},run=True)
    print str(ret)
    ret = ss.runSelect(table='deposition', select=['dep_set_id', 'depPW'], where={"dep_set_id": "'" + depid + "'"})
    print str(ret)
#   ret = ss.runInsertUpdate(table='deposition',depID='D_1000200025',data={"depPW":"'123456'"},run=True)
    print str(ret)
    ret = ss.runSelect(table='deposition', select=['dep_set_id', 'depPW'], where={"dep_set_id": "'" + depid + "'"})
    print str(ret)

    print "finished"

if __name__ == "__main__":

    main(sys.argv[1:])
